#!/usr/bin/perl -w
#
#    SaintMap v2.1 - Visual configuration tool for NetSaint
#
#    Copyright (C) 2000 David Kmoch <David.Kmoch@vslib.cz>
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

##################################################
#
# Path for directory with images
#
  my $imgpath="./images/";
#
##################################################
#
# Background image path. Use only GIF images.
# Set to empty string if you do not want any
# background image
#
# my $bg_image="somemap.gif";
  my $bg_image="";
#
##################################################
#
# Default path to your cfg files
#
#  my $cfgpath="/usr/src/netsaint";
  my $cfgpath="$ENV{HOME}";
#
#
#
##################################################

##################################################
#       NO CHANGES NEEDED BELOW                  #
##################################################

use strict;
use Tk;
use Tk::FileDialog;

my $VERSION="2.1";
my %object;
my %objprop;
my @images;
my($main,$menubar,$canvas,$dlg);
my($canvw,$canvh)=(2048,1024);
my($xpos,$ypos)=(0,0);
my $change;

my $image_format;

&init_gui;

MainLoop;

########################### SUBS #################################

sub init_gui {	# Initialize main widget, menubar and canvas
  $main=MainWindow->new(-title=>'SaintMap');
  $main->geometry(&mw_size);
  $main->protocol('WM_DELETE_WINDOW',\&quit_it);
  &create_menubar;
  &prepare;
}

###########
sub prepare { # Additional initialization, used by &close
  undef %object;
  undef %objprop;
  $change=0;
  &create_canvas;
  &read_images;
  &bgimage;
  $dlg=$main->FileDialog(-Title => 'Select hostsdefault.cfg file to read',
                         -Create => 0,
			 -FPat => '*cfg',
			 -Path => $cfgpath,
			);
}
###########
sub create_menubar {	# Create menubar, specify callbacks
  my($menuframe,$coordframe,$gridb,$aboutb,$xlabel,$ylabel,$xvalue,$yvalue);
  $menuframe=$main->Frame(qw/-relief raised -borderwidth 2/)
                     ->pack(qw/-side top -anchor n -expand 1 -fill x/);

  $menubar=$menuframe->Menubutton(qw/-text File -underline 0 -menuitems/ =>
   [
    [ Button => 'Open', -command => [\&open_dialog] ],
    [ Button => 'Save', -command => [\&save_dialog],
                        -state=>'disabled'          ],
    [ Button => 'Close',-command => [\&close_current],
                        -state=>'disabled'          ],
    [ Button => 'Quit', -command => [\&quit_it]     ],
   ])->pack(-side => 'left');

  $gridb=$menuframe->Button(-text=>'Grid',
                            -relief=>'flat',
			    -command=>\&grid,
			   )
                   ->pack(-side,'left');
  $coordframe=$menuframe->Frame(-relief,'groove',-borderwidth,2)
                        ->pack(-side=>'right');
  $aboutb=$menuframe->Button(-text=>'About',
                             -relief=>'flat',
                             -command=>[\&about],
			    )
	            ->pack(-side=>'right');
  $xlabel=$coordframe->Label(-text => ' X ')
                     ->pack(-side => 'left');
  $xvalue=$coordframe->Label(-background=>'white',
                             -takefocus=>0,
			     -textvariable=>\$xpos,
			     -width=>4,
			     )
                     ->pack(-side => 'left');
  $ylabel=$coordframe->Label(-text => ' Y ')->pack(-side => 'left');
  $yvalue=$coordframe->Label(-background=>'white',
                             -takefocus=>0,
			     -textvariable=>\$ypos,
			     -width=>4,
			     )
                     ->pack(-side => 'left');
}

############
sub create_canvas {	# Create canvas, bind specific actions
  $main->update;
  my $c=$main->Scrolled("Canvas",-width => $main->width,
                                 -height => $main->height,
				 -scrollregion => [0,0,$canvw,$canvh],
				 -bg=>'gray80',
				 -scrollbars=>'osoe',
				 )
             ->pack(-fill => 'both', -expand => 1);
  $canvas=$c->Subwidget("canvas");
  $canvas->Tk::bind("<1>", [ \&get_pos, Ev('x'), Ev('y')]);
  $canvas->Tk::bind("<B1-Motion>", [ \&get_pos, Ev('x'), Ev('y')]);
  $canvas->bind("moveable","<1>", [ \&drag_start, Ev('x'), Ev('y') ]);
  $canvas->bind("moveable","<B1-Motion>", [ \&drag_it, Ev('x'), Ev('y') ]);
  $canvas->bind("moveable","<Any-Enter>", [ \&highlight,Ev('T')]);
  $canvas->bind("moveable","<Any-Leave>", [ \&highlight,Ev('T')]);
  $canvas->bind("moveable","<3>", [ \&change_type,Ev('x'), Ev('y')]);
}
############
sub bgimage { # If defined, show background image
  if(defined $bg_image) {
    my $img=$main->Photo(-file=>$bg_image);
    $canvas->createImage(0,0,-image=>$img,-anchor=>'nw',-tags=>["bgimg"]);
  }
}
############
sub grid { # Show/hide grid
  my($i);
  my $step=50;
  if($canvas->find("withtag","grid")) {
    $canvas->delete("grid");
  } else {
    for ($i=$step;$i<$canvw;$i+=$step) {
      $canvas->createLine($i,0,$i,$canvh,-fill=>'gray70',-tags,["grid"]);
    }
    for ($i=$step;$i<$canvh;$i+=$step) {
      $canvas->createLine(0,$i,$canvw,$i,-fill=>'gray70',-tags,["grid"]);
    }
    $canvas->lower("grid","all");
    if(defined $bg_image) {$canvas->raise("grid","bgimg")}
  }
}
############
sub get_pos {	# Callback for position indicators
  my ($obj,$x,$y)=@_;
  $xpos=$canvas->canvasx($x);
  $ypos=$canvas->canvasy($y);
}

###########
sub mw_size { # Set MainWindow size
  my $size=0.9;
  my($dx,$dy)=($main->screenwidth,$main->screenheight);
  $dx = $dx > 1024 ? 1024 : $dx;
  $dy = $dy > 768 ? 768 : $dy;
  return int($size*$dx)."x".int($size*$dy)
}
############
sub open_dialog { # Show open file dialog
  my($fname);
  if($fname=$dlg->Show) {
    &read_hosts($fname);
    $dlg->configure(-Title => 'Select hostextinfodefault.cfg file to read',
                    -FPat => '*cfg',
                    -File => '',
		    );
    if($fname=$dlg->Show) {
      &read_nscgi($fname);
    }

### NEW ###

#    &update_objects;
#    &create_objects;


  }
}
############
sub save_dialog { # Show save file dialog
  my($fname);
    $dlg->configure(-Title => 'Save as...',
                    -FPat => '*cfg',
                    -File => '',
		    -Create => 1,
		    );
    if($fname=$dlg->Show) {
      if (!(-e $fname) || ((-e $fname) && (&overwrite($fname)))) {
        &save_file($fname);
      }
    return $fname;
    }
}
############
sub save_file { # Save %object records into $file
  my($file)=@_;
  my($obj,$img1,$img2);
  unless(open(SAVE,">$file")) {
    warn "$file : $!\n";
    return
  }
  for $obj (keys %object) {

    printf SAVE "define hostextinfo {\n";
    printf SAVE "host_name        %s\n", $obj;
    if($object{$obj}{"Notes"} ne "")
    {
    printf SAVE "notes_url        %s\n", $object{$obj}{"Notes"};
    }
    if($object{$obj}{"Img"} ne "")
    {
    printf SAVE "icon_image       %s\n", $object{$obj}{"Img"};
    }
    if($object{$obj}{"Alt"} ne "")
    {
    printf SAVE "icon_image_alt   %s\n", $object{$obj}{"Alt"};
    }
    if($object{$obj}{"VRML"} ne "")
    {
    printf SAVE "vrml_image       %s\n", $object{$obj}{"VRML"};
    }
    if($object{$obj}{"GD2"} ne "")
    {
    printf SAVE "statusmap_image  %s\n", $object{$obj}{"GD2"};
    }
    printf SAVE "2d_coords        %s,%s\n", $object{$obj}{"X2d"}, $object{$obj}{"Y2d"};
    if($object{$obj}{"X3d"} ne "")
    {
    printf SAVE "3d_coords        %s,%s,%s\n", $object{$obj}{"X3d"}, $object{$obj}{"Y3d"}, $object{$obj}{"Z3d"};
    }
    printf SAVE "}\n\n";

  }
  close(SAVE);
  $change=0;
}
############
sub overwrite { # Toplevel overwrite dialog, returns 0 or 1
  my($file)=@_;
  my($over,$map,$text,$butf,$textf,$yes,$no,$retval);
  $over=$main->Toplevel(-title=>'Overwrite file?');
  $over->geometry("+".int($main->width/2)."+".int($main->height/2));
  $over->transient($main);
  $over->grab;
  $textf=$over->Frame->pack;
  $map=$textf->Label(-bitmap=>'error')->pack(-side,'left',-padx,5,-pady,5);
  $text=$textf->Label(-text=>"File already exists!\nOverwrite?")->pack(-side=>'left');
  $butf=$over->Frame->pack;
  $yes=$butf->Button(-text=>"Yes",-command=>sub{$retval=1})->pack(-side,'left');
  $no=$butf->Button(-text=>"No",-command=>sub{$retval=0})->pack(-side,'left');
  $over->waitVariable(\$retval);
  $over->grabRelease;
  $over->destroy;
  return $retval;
}
############
sub ask_save { # Toplevel ask dialog, returns 0 or 1
  my($over,$map,$text,$butf,$textf,$yes,$no,$retval);
  $over=$main->Toplevel(-title=>'Save changes?');
  $over->geometry("+".int($main->width/2)."+".int($main->height/2));
  $over->transient($main);
  $over->grab;
  $textf=$over->Frame->pack;
  $map=$textf->Label(-bitmap=>'warning')->pack(-side,'left',-padx,5,-pady,5);
  $text=$textf->Label(-text=>"Save changes?")->pack(-side=>'left');
  $butf=$over->Frame->pack;
  $yes=$butf->Button(-text=>"Yes",-command=>sub{$retval=1})->pack(-side,'left');
  $no=$butf->Button(-text=>"No",-command=>sub{$retval=0})->pack(-side,'left');
  $over->waitVariable(\$retval);
  $over->grabRelease;
  $over->destroy;
  return $retval;
}
############
sub read_hosts {# Actual reading of file hosts, parsing into %object hash
  my $file = shift;
  my($line,$host,$parent);
  open(HOST,$file) or die("$file : $!");
  while($line=<HOST>) {
    next unless(($host,$parent)=$line=~/^host\[([^\]]+).*?;.*?;([^;]+)?/);
    $object{$host}->{"Parent"}=$parent;
  }
  close(HOST);
}
############
sub read_nscgi {  # Actual nscgi reading, completing records in %object
  my($file)=@_;
  my($line,$host,$img,$notes,$vrml,$gd2,$alt);
  my($xy,$xyz,$x_2d,$y_2d,$x_3d,$y_3d,$z_3d);
  open(NSCGI,$file) or die("$file : $!");
  while($line=<NSCGI>) {
    next unless ($host,$line) = $line =~ /^hostextinfo\[([^\]]+)\]=(.*)/;
    ($notes,$img,$vrml,$gd2,$alt,$xy,$xyz)=split(';',$line);
    ($x_2d,$y_2d)=split(',',$xy);
    ($x_3d,$y_3d,$z_3d)=split(',',$xyz);
    if(exists $object{$host}) {
      $object{$host}->{"Img"}=$img;
      $object{$host}->{"Notes"}=$notes;
      $object{$host}->{"Img"}=$img;
      $object{$host}->{"VRML"}=$vrml;
      $object{$host}->{"GD2"}=$gd2;
      $object{$host}->{"Alt"}=$alt;
      $object{$host}->{"X2d"}=$x_2d;
      $object{$host}->{"Y2d"}=$y_2d;
      $object{$host}->{"X3d"}=$x_3d;
      $object{$host}->{"Y3d"}=$y_3d;
      $object{$host}->{"Z3d"}=$z_3d;
      }
    }
    close(NSCGI);
    &update_objects;
    &create_objects;
}

############
sub update_objects { # Correcting read values
  my($obj,$x,$y);
  my $counter=0;
  for $obj (keys %object) {
    unless (defined $object{$obj}->{"Y2d"}) {
      $counter+=60;
      $x=60+($counter%800);
      $y=(int($counter/800)+1)*40;
      $object{$obj}->{"X2d"}=$x;
      $object{$obj}->{"Y2d"}=$y;
    }
    unless (defined $object{$obj}->{"Img"}) {
      $object{$obj}->{Img}="unknown.gif";
    }
    for $x (qw/Notes VRML GD2 Alt X3d Y3d Z3d/) {
      unless (defined $object{$obj}->{$x}) {
	$object{$obj}->{$x}="";
      }
    }
  }
}
############
sub create_objects { # Create canvas objects stored in %object
  my($obj,$img,$lineid,$l,$t,$r,$b);
  for $obj (keys %object) {

#	$img=$main->Photo(-file,$imgpath.$object{$obj}->{"Img"});
	$img=$main->Photo(-file,$imgpath."unknown.gif");


    $object{$obj}->{"ID"}=$canvas->createImage($object{$obj}->{"X2d"},
                                               $object{$obj}->{"Y2d"},
					       -image => $img,
					       -tags => ["moveable",$obj]
					       );
    ($l,$t,$r,$b)=$canvas->bbox($object{$obj}->{"ID"});
    $object{$obj}->{LabelID}=$canvas->createText($l+20,$b+10,
                                                 -font => 'fixed',
						 -text => $obj
						 );
    if(defined($object{$obj}->{"Parent"}) && (exists $object{$object{$obj}->{"Parent"}})) {
      $lineid=$canvas->createLine($l+(int(($r-$l)/2)),
                                  $t+(int(($b-$t)/2)),
				  $object{$object{$obj}->{"Parent"}}->{"X2d"},
				  $object{$object{$obj}->{"Parent"}}->{"Y2d"},
				  -tags => ["line"]
				 );
      $object{$obj}->{LineID}->{$lineid}=$object{$obj}->{"Parent"};
      $object{$object{$obj}->{"Parent"}}->{LineID}->{$lineid}=$obj;
    }
  }
  $canvas->lower("line","moveable");
  $menubar->entryconfigure('Save',-state=>'normal');
  $menubar->entryconfigure('Close',-state=>'normal');
  $menubar->entryconfigure('Open',-state=>'disabled');
}

############
sub highlight { # Highlight object's label under the mouse cursor
  my($obj,$event)=@_;
  my($name);
  $name=($canvas->gettags($canvas->find("withtag","current")))[1];
  if($event eq 'EnterNotify') {
    $canvas->itemconfigure($object{$name}->{LabelID},-fill => 'Red');
  } else {
    $canvas->itemconfigure($object{$name}->{LabelID},-fill => 'Black');
    }
}
############
sub drag_start { # Callback for button 1 click
  my ($obj,$x,$y) = @_;
  $objprop{x}=$x;
  $objprop{y}=$y;
  $objprop{obj}=$canvas->find("withtag","current");
}
############
sub drag_it { # Callback for button 1 motion - moves object, label and lines
  my ($obj,$x,$y) = @_;
  my($dx,$dy,$name,$l1,$r1,$t1,$b1,$l2,$r2,$t2,$b2);
  $change=1 unless $change;
  $dx=($x-$objprop{x});
  $dy=($y-$objprop{y});
  $canvas->move($objprop{obj},$dx,$dy);
  $objprop{x}=$x;
  $objprop{y}=$y;
  $name=($canvas->gettags($objprop{obj}))[1];
  $canvas->move($object{$name}->{LabelID},$dx,$dy);
  ($object{$name}{"X2d"},$object{$name}{"Y2d"})=$canvas->coords($objprop{obj});
  for my $line (keys %{$object{$name}{LineID}}) {
    ($l1,$r1,$t1,$b1)=$canvas->bbox($objprop{obj});
    ($l2,$r2,$t2,$b2)=$canvas->bbox($object{$object{$name}{LineID}{$line}}{"ID"});
    $canvas->coords($line,$l1+(int(($t1-$l1)/2)),$r1+(int(($b1-$r1)/2)),$l2+(int(($t2-$l2)/2)),$r2+(int(($b2-$r2)/2)));
  }
}
############
sub change_type { # Callback for button 3 - change type of the object
  my($obj,$x,$y)=@_;
  my($id,$name,$img,$choice);
  $change=1 unless $change;
  $id=$canvas->find("withtag","current");
  $name=($canvas->gettags($id))[1];
  $img=$object{$name}{"Img"};
  $choice=&choose_type($x,$y,$img);
  $object{$name}{"Img"}=$choice;
  $img=$main->Photo(-file=>$imgpath.$choice);
  $canvas->itemconfigure($id,-image=>$img);
}
############
sub choose_type { # Change type top level widget - returns image file
  my($x,$y,$current)=@_;
  my($count,$cv,$type,$img,$i,$h,$w,$wmax,$hsum);
  my $space=4;
  $type=$main->Toplevel(-title=>"Choose type");
  $type->geometry("+$x+$y");
  $type->resizable(0,1);
  $type->transient($main);
  $x=$type->Scrolled('Canvas', -height => 300,
			       -scrollbars=>'osoe',
	            )
          ->pack(-fill => 'both', -expand => 1);
  $cv=$x->Subwidget("canvas");
  $hsum=0;
  $count=$space;
  $wmax=0;
  for $img (@images) {
    $i=$main->Photo(-file=>$imgpath.$img);
    $h=$i->height;
    $w=$i->width;
    $wmax= $w > $wmax ? $w : $wmax;
    $hsum+=$h;
    $y=$cv->Radiobutton(-image=>$i,
                        -height=>$h,
                        -value=>$img,
                        -variable=>\$current,
                        -command=>sub{$type->grabRelease;$type->destroy},
                       );
    $cv->createWindow($w,$count+int($i->height/2), -window => $y);
    $count=$count+$h+$space;
  }
  $x->configure(-width => $wmax+50,
                -scrollregion => [0,0,$wmax+50,$space+$hsum+$#images*$space]
	       );
  $type->update;
  $type->grab;
  $type->waitWindow;
  return $current;
}
############
sub close_current { # Close current map
  if ($change) {
    if(&ask_save && !&save_dialog) {
      return
    }
  }
  $menubar->entryconfigure('Save',-state=>'disabled');
  $menubar->entryconfigure('Close',-state=>'disabled');
  $menubar->entryconfigure('Open',-state=>'normal');
  $canvas->parent->destroy;
  &prepare;
}
############
sub read_images { # Read image dir into @images
  opendir(DIR, $imgpath) || die "$imgpath : $!";
  @images = readdir(DIR);
  shift @images;
  shift @images;
  closedir DIR;
}
############
sub about { # Show about
  my($label,$about,$but);
  my $MSG="Copyright (c) 2000-2001
Author: David Kmoch <David.Kmoch\@vslib.cz>\n\nThanks to:\n
Petr Adamec <Petr.Adamec\@vslib.cz> for the idea\n
Ethan Galstad <netsaint\@netsaint.org> for NetSaint\n
Adrian Pavlykevych <pam\@polynet.lviv.ua>,
Szilard Fulop <silas\@fornax.hu>,
for patches and new ideas ";
  $about=$main->Toplevel(-title => 'About SaintMap');
  $about->geometry("+".int($main->width/2)."+".int($main->height/2));
  $about->transient($main);
  $about->resizable(0,0);
  $about->grab;
  $label=$about->Label(-text=>"SaintMap v$VERSION\n\n $MSG",
                       -font => "7x13",
                      )
               ->pack(-padx => 5,
	              -pady => 5,
		      -expand => 1
		      );
  $but=$about->Button(-text=>'OK',
                      -command=>sub{$about->grabRelease;$about->destroy},
		     )
             ->pack(-pady => 10);
}
############
sub quit_it { # Obvious
  $main->destroy;
  exit;
}
